package org.fsou.namematch.slavonic;

import java.util.*;
import org.fsou.namematch.*;

/**
 * Generic matcher of slavonic names written in English.
 * Takes into account prevalent rules of slavonic name transliteration.
 * @author Alexander Yevstafyev
 * @version 1.0.0
 */
public class SlavonicMatcher implements INameMatcher {

	/**
	 * Prevents instantiation of this (unaccomplished) class.
	 */
	protected SlavonicMatcher() {
	}

	/**
	 * Matches two slavonic names written in English.
	 * Ignores all non-English letters.
	 */	
	public IMatchResult matchNames(String name1, String name2) {
		
		name1 = name1.toLowerCase();
		name2 = name2.toLowerCase();
		if (name1.equals(name2)) {
			return new MatchResult(1);
		}
		
		name1 = getSkeleton(name1);
		name2 = getSkeleton(name2);
		if (name1.equals(name2)) {
			boolean doubtful = name1.length()<=1;
			return new MatchResult(1, doubtful);
		}
		
		int hits = countHits(name1, name2);
		int length = Math.max(name1.length(), name2.length());
		
		boolean doubtful = false;
		if (length==1) {
			doubtful = true;
		}
		if (length==2 && hits==1) {
			doubtful = true;
		}
		
		return new MatchResult(((float)hits)/length, doubtful);
	}

	/**
	 * Extracts so called skeleton of name. Skeleton is formed by name's consonants
	 * (vowels are ejected). Additionally, some consonant replacement performed.
	 * @param name Original name in lower case
	 * @return Name skeleton
	 */	
	protected String getSkeleton(String name) {

		// Filter illegal letters.		
		StringBuffer filteredName = new StringBuffer();
		for (int i=0; i<name.length(); i++) {
			char letter = name.charAt(i);
			if (letter<'a'|| letter>'z') {
				continue;
			}
			filteredName.append(letter);
		}
		name = filteredName.toString();
		
		name = replaceCombinations(name);

		// Filter vowels.
		StringBuffer skeleton = new StringBuffer();
		for (int i=0; i<name.length(); i++) {
			char letter = name.charAt(i);
			if (!isConsonant(letter)) {
				continue;
			}
			skeleton.append(letter);
		}
		
		return skeleton.toString(); 	
	}

	/**
	 * Should replace widespread combinations of several english letters which
	 * corespond to a single slavonic one, or single english letter that corresponds
	 * to several slavonic ones, etc. 
	 * @param name Original name in lower case
	 * @return Name with replaced combinations.
	 */	
	protected String replaceCombinations(String name) {
		return name;
	}

	/**
	 * Determines if given letter is consonant.
	 * Subclasses should supplement this method taking into account replacements of
	 * combinations done by them.
	 * @param letter Letter in lower case
	 * @return If letter is consonant
	 */	
	protected boolean isConsonant(char letter) {
		return letter!='a'
		    && letter!='e'
		    && letter!='i'
		    && letter!='o'
		    && letter!='u'
		    && letter!='y';
	}

	/**
	 * Counts size of a most long mutual discontinuous sequence of consonants
	 * contained in both name skeletons.
	 * @param skeleton1 First name skeleton
	 * @param skeleton2 Second name skeleton
	 * @return Size of most long mutual sequence
	 */	
	protected int countHits(String skeleton1, String skeleton2) {
		
		int hits = 0;
		Set sequences1 = getSequences(skeleton1);
		Set sequences2 = getSequences(skeleton2);
		
		for (Iterator it=sequences1.iterator(); it.hasNext(); ) {
			String sequence = (String)it.next();
			if (sequences2.contains(sequence)) {
				int length = sequence.length();
				if (length>hits) {
					hits = length;
				}
			}
		}
		
		return hits;
	}

	/**
	 * Extracts all possible discontinuous sequences of consonants contained in
	 * name skeleton. 
	 * @param skeleton Name skeleton
	 * @return All possible discontinuous sequences of consonants as Set of Strings.
	 */	
	protected Set getSequences(String skeleton) {
		Set sequences = new HashSet();
		addSequences(sequences, "", skeleton, 0);
		return sequences;
	}
	
	/**
	 * Adds to set all sequences which can be formed by adding letters situated
	 * starting from current sequence base index, to current sequence base.
	 * @param sequences Set of sequences being formed
	 * @param sequence Current sequence base
	 * @param skeleton Original name skeleton
	 * @param from Current sequence base index
	 */
	protected void addSequences(Set sequences, String sequence, String skeleton, int from) {
		sequences.add(sequence);
		for (int i=from; i<skeleton.length(); i++) {
			addSequences(sequences, sequence+skeleton.charAt(i), skeleton, i+1);
		}
	}
}